#include <iostream>
#if __has_include(<memory_resource>)
#  include <memory_resource>
#else
#  include <experimental/memory_resource>
#endif // __has_include (<memory_resource>)
#include <coroutine>
#include <limits>
#include <optional>
#include <string_view>
#if __has_include(<source_location>)
#  include <source_location>
#endif // __has_include(<source_location>)

#if __has_include(<format>)
#  include <format>
#  define std_format(...) std::format(__VA_ARGS__)
// #  define std_print(...) std::print(__VA_ARGS__)
# define std_print(...)
#else
#  define std_format(...) ""
#  define std_print(...)
#endif

#include <stdio.h>
#ifdef __GLIBC__
#  include <gnu/libc-version.h>
#endif

#include <ranges>
#include <utility>

#if __cpp_lib_source_location
void
log(
  const std::string_view message,
  const std::source_location location = std::source_location::current())
{
  std::cout << "file: " << location.file_name()                            //
            << "(" << location.line() << ":" << location.column() << ") `" //
            << location.function_name() << "`: " << message << '\n';
}
#endif // __cpp_lib_source_location

class T
{
public:
  auto print(std::string_view message) const -> void;
  static auto prints(std::string_view message) -> void;
};

auto
T::print(std::string_view message) const -> void
{
  (void)message;
#if __cpp_lib_source_location
  constexpr auto location = std::source_location::current();
  std::cout << "file: " << location.file_name()                            //
            << "(" << location.line() << ":" << location.column() << ") `" //
            << location.function_name() << "`: " << message << '\n';
#endif
}

auto
T::prints(std::string_view message) -> void
{
#if __cpp_lib_source_location
  constexpr auto location = std::source_location::current();
  std::cout << "file: " << location.file_name()                            //
            << "(" << location.line() << ":" << location.column() << ") `" //
            << location.function_name() << "`: " << message << '\n';
#endif

  auto data = message.data();
  (void)data;

  auto message1 = message;
  (void)message1;
}

auto
main() -> int
{
  int min = std::numeric_limits<int>::min();
  std::cout << min << std::endl;

  int max = std::numeric_limits<int>::max();
  std::cout << max << std::endl;

  auto max1 = std::numeric_limits<int>::max();
  (void)max1;

  auto max2 = std::numeric_limits<unsigned int>::max();
  (void)max2;

  auto max3 = std::numeric_limits<std::uint64_t>::max();
  (void)max3;

  std::cout << "----> monadic optional\n";
  std::optional<int> i1;
  std::optional<int> i2{ 10 };

  if (i1 && i2) {
    std::cout << *i2 << std::endl;
  }
  std::cout << "monadic optional value :"
            << i2.transform([](auto v) { return v * v; })
                 .transform([](auto v) { return v + 1; })
                 .and_then([](auto v) {
                   std::cout << "\n from: " << v << "\n";
                   return std::make_optional(5);
                 })
                 .value_or(-5)
            << "\n";
  
    const auto val = i2
        .transform([](auto v) { return v * 10000; })
        .and_then([](auto v) {
            std::cout << "here : " << v << "\n";
            return std::make_optional(5);
        })
        .or_else([](){return std::make_optional(-6);})
        .value();
    std::cout << "result: " << val << "\n";

#if __cpp_lib_source_location
  const auto location = std::source_location::current();
  std::cout << "function name: " << location.function_name() << std::endl;
  std::cout << "line:          " << location.line() << std::endl;
  std::cout << "column:        " << location.column() << std::endl;
  std::cout << "file:          " << location.file_name() << std::endl;

  log("heho");
#endif // __cpp_lib_source_location

  auto i = 10;
  (void)i;

  auto a = 10uz;
  (void)a;

  auto f = 0.0;
  (void)f;

  auto f1 = 0.0f;
  (void)f1;

  std::cout << "----> Hello world and print\n";

  std::cout << "Hello, World!" << std::endl;

    std::cout << "----> source location\n";

  T t;
  t.print("Time to print");
  T::prints("Time to prints!!!");

std::cout << "-----> LIBs\n";

#ifdef __GLIBCXX__
  std::printf("GLIBCXX: %d\n", __GLIBCXX__);
#else
  puts("Not the GLIBCXX Library");
#endif
#ifdef __GLIBCPP__
  std::printf("GLIBCPP: %d\n", __GLIBCPP__);
#else
  puts("Not __GLIBCPP__ Library");
#endif

#ifdef __GLIBC__
  printf("GNU libc compile-time version: %u.%u\n", __GLIBC__, __GLIBC_MINOR__);
  printf("GNU libc runtime version:      %s\n", gnu_get_libc_version());
#else
  puts("Not the GNU C Library");
#endif

#ifdef _LIBCPP_VERSION
  std::printf("_LIBCPP_VERSION: %d\n", _LIBCPP_VERSION);
#else
  puts("Not _LIBCPP_VERSION Library");
#endif

  std::cout << "----> Ranges\n";
  auto const ints = { 0, 1, 2, 3, 4, 5 };
  auto even = [](int i) { return 0 == i % 2; };
  auto square = [](int i) { return i * i; };

  // "pipe" syntax of composing the views:
  for (int i :
       ints | std::views::filter(even) | std::views::transform(square)) {
#if __has_include(<format>)
    std::format("----> std::format - where am i? {}", "\n");
#endif
    std::cout << i << ' ';
  }

  std::cout << '\n';

  // a traditional "functional" composing syntax:
  for (int i : std::views::transform(std::views::filter(ints, even), square))
  {
    std::cout << i << ' ';
  }
  std::cout << "\n";

  std::cout << "-----> std::format\n";

  std::cout << std_format("help {}\n", ++i);
  std_print(stdout, "help {}\n", ++i);

  std::cout << "-----> std::unreachable\n";

  if (true) {
    return 0;
  } else {
    std::unreachable();
    return 1;
  }
}
